/* --------------------------------------------------------------------------
$Id: read_hlhdf.c,v 1.1.1.1 2001/12/14 12:02:40 ahenja Exp $
Program:        

Description:    Used for reading an HDF5 file and building a node tree
		of NameList types.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: read_hlhdf.c,v $
Revision 1.1.1.1  2001/12/14 12:02:40  ahenja
First CVS version of hlhdf

Revision 1.11  2001/11/13 15:18:44  adybbroe
fixed for bug in createCompoundObject.

Revision 1.10  2001/11/12 10:24:37  adybbroe
corrected for new type H5T_ARRAY.

Revision 1.9  2001/11/08 08:54:35  ahenja
Removed comments, etc. seems like Adam has solved all problems related to HDF5.4.2, I don't know if it works with HDF5.4.1

Revision 1.8  2001/11/08 08:15:24  adybbroe
Forced checkin since ahenja has to access it.

Revision 1.7  2001/04/26 14:16:55  adybbroe
Solved for version 1.4 incompatibility.

Revision 1.6  2000/03/22 08:36:09  ahenja
Updated with spelling corrections and header information.

-----------------------------------------------------------------------------*/
#include "read_hlhdf.h"

/*Forward declarations*/
static herr_t groupIterator(hid_t group,const char* name, void* op_data);
static herr_t groupAttributeIterator(hid_t loc_id, const char* name, void* op_data);
static herr_t datasetAttributeIterator(hid_t loc_id, const char* name, void* op_data);
static NameListAttribute_t* readAttribute(hid_t obj, const char* name);
static herr_t readDataset(NameListGroup_t* group,hid_t gid, const char* name, const char* dsetName);
static NameListType_t* buildNameListType(hid_t type,NameListGroup_t* pGrp,const char* name);
static NameListType_t* buildCompoundDataType(NameListGroup_t* pGrp, hid_t type, const char* name,int isNamed);
static NameListType_t* createCompoundObject(hid_t type,NameListGroup_t* pGrp);

NameListGroup_t* readHlHdfFileFrom(const char* filename, const char* from)
{
   hid_t file_id,gid;
   NameListGroup_t* rootGrp=NULL;

   if((file_id=openHlHdfFile(filename,"r"))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open file %s\n",filename);
      return NULL;
   }

   rootGrp = newGroup(NULL,from);

   if((gid=H5Gopen(file_id,from))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open file from %s",from);
      goto fail;
   }
   H5Aiterate(gid,NULL,groupAttributeIterator,rootGrp);
   H5Gclose(gid);

   if(H5Giterate(file_id,from,NULL,groupIterator,rootGrp)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not iterate over groups\n");
      goto fail;
   }
   H5Fclose(file_id);
   return rootGrp;
fail:
   H5Fclose(file_id);
   read_hlhdf_free(rootGrp);
   return NULL;
}

NameListGroup_t* readHlHdfFile(const char* filename)
{
   return readHlHdfFileFrom(filename,"/");
}

#ifdef KALLE
NameListGroup_t* readHlHdfFile(const char* filename)
{
   hid_t file_id,gid;
   NameListGroup_t* rootGrp=NULL;

   if((file_id=openHlHdfFile(filename,"r"))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open file %s\n",filename);
      return NULL;
   }

   rootGrp = newGroup(NULL,"/");

   gid=H5Gopen(file_id,"/");
   H5Aiterate(gid,NULL,groupAttributeIterator,rootGrp);
   H5Gclose(gid);

   H5Giterate(file_id,"/",NULL,groupIterator,rootGrp);

   H5Fclose(file_id);

   return rootGrp;
}
#endif


/************************************************
 * read_hlhdf_free
 * This function hopefully knows how to free all memory
 * that has been allocated by the readHlhdfFile
 ***********************************************/
void read_hlhdf_free(NameListGroup_t* group)
{
   int i=0;

   if(group->attrs) {
      for(i=0;i<group->noAttrs;i++) {
	 freeAttribute(group->attrs[i]);
      }
      free(group->attrs);
   }

   if(group->datasets) {
      for(i=0;i<group->noDatasets;i++) {
	 freeInternalDataset(&group->datasets[i]);
      }
      free(group->datasets);
   }

   for(i=0;i<group->noGroups;i++) {
      read_hlhdf_free(&group->groups[i]);
   }

   if(group->groups)
      free(group->groups);

   if(group->thisIsRootGrp) {
      for(i=0;i<group->noGlobalTypes;i++) {
	 freeNameListType(group->globalTypes[i]);
      }
      free(group->globalTypes);
   }

   if(group->name)
      free(group->name);

   free(group->localTypes);
   if(group->thisIsRootGrp)
      free(group);
}

/*EVERYTHING BELOW THIS POINT IS FOR INTERNAL USE ONLY, I.E. STATIC*/
static NameListType_t* createCompoundObject(hid_t type,NameListGroup_t* pGrp)
{
   int nmembers;
   int ndims;
   size_t* dims=NULL;
   hsize_t* dims_h514=NULL;
   hid_t mtype;
   int i,j;
   char* fname=NULL;
   NameListType_t* newType=NULL;

   newType = newNameListType();
   newType->isCompound=1;

   newType->typeSize = H5Tget_size(type);

   nmembers=H5Tget_nmembers(type);
   for(i=0;i<nmembers;i++) {
      NameListType_t* tmpType=NULL;
      CompoundAttributeDef_t* compoundAttr=NULL;
      size_t offset = H5Tget_member_offset(type,i);
      fname = H5Tget_member_name(type,i);
      mtype = H5Tget_member_type(type,i);

#ifdef H5_VERS_MINOR
#if H5_VERS_MINOR>=4
      if (H5T_ARRAY==H5Tget_member_class(type,i)) {
	 ndims = H5Tget_array_ndims(mtype);
	 dims_h514 = (hsize_t*) malloc(sizeof(hsize_t)*ndims);
	 if(H5Tget_array_dims(mtype,dims_h514,NULL) < 0) {
	    fprintf(stderr,"read_hlhdf - Failed to retrieve array dimensions\n");
	 }
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);	 
	 for(j=0; j < ndims; j++) {
	    dims[j] = (size_t) dims_h514[j];
	 }
      }
      else {
	 ndims=1;
	 dims = (size_t*)malloc(sizeof(size_t)*ndims);
	 dims[0]=1;
      }	       
#else
      ndims = H5Tget_member_dims(type,i,NULL,NULL);
      dims = malloc(sizeof(size_t)*ndims);
      ndims = H5Tget_member_dims(type,i,dims,NULL);
      /*
	Incompatible with version 1.4 and higher. Adam Dybbroe, 2001-04-26
      */
#endif
#else
      if (H5T_ARRAY==H5Tget_member_class(type,i)) {
	 ndims = H5Tget_array_ndims(mtype);
	 dims_h514 = malloc(sizeof(hsize_t)*ndims);
	 if(H5Tget_array_dims(mtype,dims_h514,NULL) < 0) {
	    fprintf(stderr,"read_hlhdf (no H5_VERS_MINOR) - Failed to retrieve array dimensions\n");
	 }	 
	 dims = malloc(sizeof(size_t)*ndims);	 
	 for(j=0; j < ndims; j++) {
	    dims[j] = (size_t) dims_h514[j];
	 }
      }
      else {
	 ndims=1;
	 dims = malloc(sizeof(size_t)*ndims);
	 dims[0]=1;
      }	       
#endif

      tmpType = buildNameListType(mtype,pGrp->rootGroup,fname);
      
      compoundAttr = createCompoundFromType(tmpType,fname);

      compoundAttr->offset=offset;

      freeNameListType(tmpType);

      compoundAttr->ndims=ndims;

      for(j=0;j<ndims;j++) {
	 compoundAttr->dims[j]=dims[j];
      }

      if(addCompoundAttributeToType(newType,compoundAttr)<0) {
	 fprintf(stderr,"Failed to add compound attribute to type\n");
      }

      if(mtype>=0) H5Tclose(mtype);
      if(fname) free(fname);
      if(dims) {
	 free(dims);
	 dims=NULL;
      }
      if(dims_h514) {
	 free(dims_h514);
	 dims_h514 = NULL;
      }
   }

   return newType;
}

static NameListType_t* buildCompoundDataType(NameListGroup_t* pGrp, hid_t type, const char* name,int isNamed)
{
   int nmembers;
   int ndims;
   size_t* dims=NULL;
   hid_t mtype;
   int i,j;
   char* fname;
   H5G_stat_t statbuf;
   NameListType_t* newType=NULL;

   if(H5Tcommitted(type)>0) {
      H5Gget_objinfo(type,".",TRUE,&statbuf);

      i=doesTypeExistInGlobalGroup(pGrp->rootGroup,statbuf.objno);

      if(i<0) {
	 /*Nope the type does not exist, then I better create it*/
	 char typeName[1024];
	 if(!isNamed) {
	    sprintf(typeName,"/#%lu:%lu",statbuf.objno[0],statbuf.objno[1]);
	 } else {
	    sprintf(typeName,"%s",name);
	 }
	 newType = createCompoundObject(type,pGrp->rootGroup);
	 newType->name=strdup(typeName);
	 newType->isNamed=isNamed;
	 newType->objno[0]=statbuf.objno[0];
	 newType->objno[1]=statbuf.objno[1];

	 if(addTypeToGlobalGroup(pGrp->rootGroup,newType)<0) {
	    fprintf(stderr,"Failed to add type definition to root group\n");
	 }

	 if(isNamed) {
	    if(addTypeToLocalGroup(pGrp,newType)<0) {
	       fprintf(stderr,"Failed to add local type definition to local group\n");
	    }
	 } else {
	    if(addTypeToLocalGroup(pGrp->rootGroup,newType)<0) {
	       fprintf(stderr,"Failed to add local type definition to local root group\n");
	    }
	 }
      } else {
	 newType=pGrp->rootGroup->globalTypes[i];
	 if(isNamed) {
	    newType->isNamed=isNamed;
	    if(newType->name)
	       free(newType->name);
	    /*If the datatype already exists, then change the name*/
	    newType->name=strdup(name);
	 }

	 if(isNamed) {
	    /*If this is the case, then I must transfer the group from the root groups localTypes
	     *to the subgroups localTypes
	     */
	    NameListType_t* tmpGrp=NULL;
	    int rootLocalIdx=-1;
	    int localIdx = -1;

	    rootLocalIdx = doesTypeExistInLocalGroup(pGrp->rootGroup,newType->objno);
	    if(rootLocalIdx>=0) {
	       /*Jupp, then I better remove the reference from root groups local list,
		*and transfer this pointer to the current groups local list*/
	       tmpGrp = removeTypeFromLocalGroup(pGrp->rootGroup,newType->objno);
	       /*And add this pointer to the local groups type list*/
	       if(addTypeToLocalGroup(pGrp,tmpGrp)<0) {
		  fprintf(stderr,"Failed to transfer the type from the root group to the local group\n");
	       }
	    } else {
	       /*Nope the type did not exist there, then try this group also*/
	       localIdx = doesTypeExistInLocalGroup(pGrp,newType->objno);
	       if(localIdx>=0) {
		  /*It already exists here, do nothing about it then*/
	       } else {
		  /*Weehaa, just add the local group type*/
		  if(addTypeToLocalGroup(pGrp,newType)<0) {
		     fprintf(stderr,"Failed to transfer the type from the root group to the local group\n");
		  }
	       }
	    }
	 }
      }

   } else { /*This is a local datatype, do not bother to add the datatype to any group,
	     *since it hopefully does not exist anywhere else*/
      newType=createCompoundObject(type,pGrp);
   }

   return newType;
}

static NameListType_t* buildNameListType(hid_t type,NameListGroup_t* pGrp,const char* name)
{
   NameListType_t* retv=NULL;
   H5G_stat_t statbuf;

   switch(H5Tget_class(type)) {
   case H5T_INTEGER:
   case H5T_FLOAT:
      /*Handle standard types here*/
      retv=newNameListType();
      retv->referenceId = getFixedType(type);
      retv->typeSize=H5Tget_size(retv->referenceId);
      retv->ieee_type=getTypeNameString(type);
      retv->format_type=getFormatNameString(retv->referenceId);
      H5Tclose(retv->referenceId);
      break;
   case H5T_STRING:
      /*Handle strings here*/
      retv=newNameListType();
      retv->referenceId = getFixedType(type);
      retv->isString=1;
      retv->typeSize=H5Tget_size(retv->referenceId);
      retv->ieee_type=getTypeNameString(type);
      retv->format_type=getFormatNameString(retv->referenceId);
      retv->strpad=getStringPadName(retv->referenceId);
      retv->strcset=getStringCsetName(retv->referenceId);
      retv->strctype=getStringCtypeName(retv->referenceId);
      H5Tclose(retv->referenceId);
      break;
   case H5T_COMPOUND:
      retv=buildCompoundDataType(pGrp->rootGroup,type,name,0);
      break;
#ifdef H5_VERS_MINOR
#if H5_VERS_MINOR >= 4
   case H5T_ARRAY:
      /*Handle array type here - Adam & Anders, 2001-11-12*/
      retv=newNameListType();
      retv->referenceId = H5Tcopy(type);
      retv->typeSize=H5Tget_size(retv->referenceId);
      retv->ieee_type=getTypeNameString(type);
      retv->format_type=getFormatNameString(retv->referenceId);
      H5Tclose(retv->referenceId);
      break;
#endif
#endif
   case H5T_ENUM:
   case H5T_TIME:
   case H5T_BITFIELD:
   case H5T_OPAQUE:
   case H5T_REFERENCE:
      printf("Unsupported types for the moment (%s)\n",name);
      break;
   default:
      printf("HEY WHAT IS THIS?? NOT A DEFINED TYPE\n");
      break;
   }

   return retv;
}

static herr_t readDataset(NameListGroup_t* group,hid_t gid, const char* name, const char* dsetName)
{
   herr_t status=0;
   hid_t type=-1,obj=-1;
   NameListDataset_t* dset=NULL;
   H5G_stat_t statbuf;

   if((obj=H5Dopen(gid,name))>=0) {
      type=H5Dget_type(obj);
      if(type<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to get type from dataset\n");
	 status=-1;
	 goto done;
      } else {
	 size_t size;
	 int ndims;
	 hsize_t* all_dims=NULL;
	 hsize_t* max_dims=NULL;
	 hsize_t npoints;
	 int j;
	 hid_t f_space=-1;
	 hid_t mtype=-1;
	 hid_t databuff_space;
	 unsigned char* dataptr=NULL;

	 size=H5Tget_size(type);
	 if((f_space=H5Dget_space(obj))>=0) {
	    ndims=H5Sget_simple_extent_ndims(f_space);
	    npoints=H5Sget_simple_extent_npoints(f_space);
	    all_dims=malloc(sizeof(hsize_t)*ndims);
	    max_dims=malloc(sizeof(hsize_t)*ndims);
	    if(H5Sget_simple_extent_dims(f_space,all_dims,max_dims)!=ndims) {
	       fprintf(stderr,"Could not get dimensions\n");
	    }

	    mtype=getFixedType(type);

	    if(H5Sis_simple(f_space)>=0) {
	       dataptr=malloc(H5Tget_size(mtype)*npoints);
	       databuff_space=H5Screate_simple(1,&npoints,NULL);

	       H5Sselect_all(f_space);
	       H5Sselect_all(databuff_space);
	       if(H5Dread(obj,mtype,databuff_space,f_space,H5P_DEFAULT,dataptr)<0) {
		  fprintf(stderr,"Failed to read dataset\n");
		  status=-1;
	       }
	       H5Sclose(databuff_space);
	    } else {
	       fprintf(stderr,"Dataspace was not simple, this is not supported\n");
	       status=-1;
	    }
	    H5Sclose(f_space);

	    if(status==0) {
	       dset=newDataset(group,dsetName);

	       dset->datatype=buildNameListType(type,group->rootGroup,dsetName);

	       /*This is a small fix, so that the offsets are adjusted according to
		*the fixed type offset, instead of the original type offset*/
	       if(H5Tget_class(type)==H5T_COMPOUND) {
		  /*Just adjust the offsets, so that they point at the correct location*/
		  int iii,jjj;
		  int nomemb=H5Tget_nmembers(mtype);
		  char* fname;
		  int gotit;
		  for(iii=0;iii<nomemb;iii++) {
		     fname=H5Tget_member_name(mtype,iii);
		     gotit=0;
		     for(jjj=0; gotit==0 && jjj<dset->datatype->noAttrs;jjj++) {
			if(strcmp(fname,dset->datatype->attrs[jjj]->name)==0) {
			   dset->datatype->attrs[jjj]->offset=H5Tget_member_offset(mtype,iii);
			   gotit=1;
			}
		     }
		     free(fname);
		  }
	       }
	       dset->ndims=ndims;
	       dset->dims=all_dims;
	       dset->data=dataptr;	       
	    } else {
	       if(dataptr)
		  free(dataptr);
	       if(all_dims)
		  free(all_dims);
	    }
	    if(max_dims)
	       free(max_dims);
	    H5Tclose(mtype);

	 }
      }
      if(dset==NULL) {
	 dset=newDataset(group,dsetName);
      }
      H5Aiterate(obj,NULL,datasetAttributeIterator,dset);
   }

 done:
   if(obj>=0)
      H5Dclose(obj);
   if(type>=0)
      H5Tclose(type);
   return status;
}

static NameListAttribute_t* readAttribute(hid_t obj, const char* name)
{
   hid_t type=-1, mtype=-1;
   hid_t f_space=-1;
   size_t size;
   hsize_t* all_dims=NULL;
   hsize_t npoints;
   int i,j;
   herr_t status=1;
   NameListAttribute_t* attr=NULL;
   int ndims;

   type=H5Aget_type(obj);
   if(type<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not get attribute type\n");
      status=-1;
      goto done;
   }

   if(H5Tcommitted(type)>0) {
      fprintf(stderr,"Attribute is a named, not supported right now\n");
      status=0;
      goto done;
   } else {
      mtype=getFixedType(type);
      if(mtype<0) {
	 status=-1;
	 goto done;
      }
      size=H5Tget_size(type);
      if((f_space=H5Aget_space(obj))>=0) {
	 ndims=H5Sget_simple_extent_ndims(f_space);
	 npoints=H5Sget_simple_extent_npoints(f_space);
	 all_dims=malloc(sizeof(hsize_t)*ndims);
	 H5Sget_simple_extent_dims(f_space,all_dims,NULL);

	 if(H5Sis_simple(f_space)>=0) {
	    attr=newAttribute(name);
	    attr->dims=malloc(sizeof(hsize_t)*ndims);
	    for(j=0;j<ndims;j++) {
	       attr->dims[j]=all_dims[j];
	    }
	    attr->ndims=ndims;
	    attr->data=malloc(H5Tget_size(mtype)*npoints);

	    if(H5Aread(obj,mtype,attr->data)<0) {
	       if(_debug_hdf)
		  fprintf(stderr,"Failed to read data from attribute\n");
	       freeAttribute(attr);
	       attr=NULL;
	       goto done;
	    }
	 } else {
	    if(_debug_hdf)
	       fprintf(stderr,"Dataspace was not simple\n");
	    goto done;
	 }
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not get dataspace\n");
	 goto done;
      }
   }
 done:
   if(type!=-1)
      H5Tclose(type);
   if(mtype!=-1)
      H5Tclose(mtype);
   if(f_space!=-1)
      H5Sclose(f_space);
   if(all_dims)
      free(all_dims);

   return attr;
}

static herr_t groupIterator(hid_t group,const char* name, void* op_data)
{
   hid_t obj,type;
   H5G_stat_t statbuf;
   hid_t mtype;
   char* tmp;
   herr_t status=0;
   NameListGroup_t* childGroup=NULL;
   NameListGroup_t* groupList=(NameListGroup_t*)op_data;
   NameListType_t* globType;
   
   H5Gget_objinfo(group,name,0,&statbuf);

   tmp=malloc((strlen(groupList->name)+strlen(name)+2)*sizeof(char));
   strcpy(tmp,groupList->name);
   strcat(tmp,name);

   switch(statbuf.type) {
   case H5G_GROUP:
      if((obj=H5Gopen(group,name))>=0) {
	 strcat(tmp,"/");
	 childGroup=newGroup(groupList,tmp);
	 H5Aiterate(obj,NULL,groupAttributeIterator,childGroup);
	 H5Giterate(obj,".",NULL,groupIterator,childGroup);
	 H5Gclose(obj);
      }
      break;
   case H5G_DATASET:
      if(readDataset(groupList,group,name,tmp)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to handle dataset %s\n",tmp);
	 status=-1;
	 }
      break;
   case H5G_TYPE:
      type = H5Topen(group,name);
      if(H5Tget_class(type)==H5T_COMPOUND) {
	 globType = buildCompoundDataType(groupList,type,tmp,1);
	 if(!globType) {
	    if(_debug_hdf)
	       fprintf(stderr,"Failed to fetch compound data type %s\n",tmp);
	    status=-1;
	 }
      }
      H5Tclose(type);
      break;
   default:
      if(_debug_hdf)
	 fprintf(stderr,"Undefined type for %s\n",name);
      break;
   }

   free(tmp);

   return status;
}

static herr_t groupAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   NameListGroup_t* grp = (NameListGroup_t*)op_data;
   herr_t status=0;
   hid_t obj;
   hid_t type;
   NameListType_t* nltype=NULL;
   NameListAttribute_t* nlattr=NULL;
   char* tmp=NULL; 

   obj=H5Aopen_name(loc_id,name);
   if(obj<0) {
      return 0;
   }

   type=H5Aget_type(obj);
   if(type<0) {
      H5Aclose(obj);
      if(_debug_hdf)
	 fprintf(stderr,"Failed to get type\n");
      return 0;
   }

   tmp=malloc((strlen(grp->name)+strlen(name)+2)*sizeof(char));
   strcpy(tmp,grp->name);
   strcat(tmp,name);

   nltype=buildNameListType(type,grp->rootGroup,tmp);
   nlattr=readAttribute(obj,tmp);
   nlattr->attrtype=nltype;

   if(nltype) {
      addAttributeToGroup(grp,nlattr);
   } else {
      fprintf(stderr,"Unsupported type fo attribute\n");
   }

   H5Tclose(type);
   H5Aclose(obj);

   free(tmp);

   return status;
}

static herr_t datasetAttributeIterator(hid_t loc_id, const char* name, void* op_data)
{
   NameListDataset_t* dset = (NameListDataset_t*)op_data;
   herr_t status=0;
   hid_t obj;
   hid_t type;
   NameListType_t* nltype=NULL;
   NameListAttribute_t* nlattr=NULL;
   char* tmp=NULL; 

   obj=H5Aopen_name(loc_id,name);
   if(obj<0) {
      return 0;
   }

   type=H5Aget_type(obj);
   if(type<0) {
      H5Aclose(obj);
      if(_debug_hdf)
	 fprintf(stderr,"Failed to get type\n");
      return 0;
   }

   tmp=malloc((strlen(dset->name)+strlen(name)+2)*sizeof(char));
   strcpy(tmp,dset->name);
   strcat(tmp,"/");
   strcat(tmp,name);

   nltype=buildNameListType(type,dset->rootGroup,tmp);
   nlattr=readAttribute(obj,tmp);
   nlattr->attrtype=nltype;

   if(nltype) {
      addAttributeToDataset(dset,nlattr);
   } else {
      printf("Unsupported type for attribute\n");
   }

   H5Tclose(type);
   H5Aclose(obj);

   free(tmp);

   return status;
}

