/* --------------------------------------------------------------------------
$Id: update_vhlhdf.c,v 1.1.1.1 2001/12/14 12:02:41 ahenja Exp $
Program:        

Description:    Functions for updating an HDF5 file with the HighLevel
		protocol.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2001

History:
$Log: update_vhlhdf.c,v $
Revision 1.1.1.1  2001/12/14 12:02:41  ahenja
First CVS version of hlhdf

Revision 1.1  2001/01/18 12:39:18  ahenja
Initial revision

-----------------------------------------------------------------------------*/
#include "update_vhlhdf.h"

int doAppendHdf5Attribute(hid_t file_id, HL_Node* parentNode, char* parentName,
			  HL_Node* childNode, char* childName)
{
   hid_t loc_id = -1;
   int status = 0;
   
   HL_Type parentType = UNDEFINED_ID;
   
   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 group since either parentName or childName is NULL\n");
      return 0;
   }

   if(strcmp(parentName,"")==0) {
      if((loc_id=H5Gopen(file_id,"/"))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not open root group when reading attr '%s'\n",
		    childName);
	 goto fail;
      }
      parentType=GROUP_ID;
   } else {
      disableErrorReporting();
      if((loc_id = H5Gopen(file_id,parentName))<0) {
	 if((loc_id = H5Dopen(file_id,parentName))>=0) {
	    parentType = DATASET_ID;
	 }
      } else {
	 parentType=GROUP_ID;
      }
      enableErrorReporting();
      if(loc_id < 0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Parent '%s' to attribute '%s' could not be opened\n",
		    parentName,childName);
	 goto fail;
      }
   }

   if(strcmp(childNode->format,DATAFORMAT_UNDEFINED)!=0 && childNode->typeId<0) {
      childNode->typeId = translateCharToDatatype(childNode->format);
   }
   if(childNode->typeId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't recognize datatype '%s' or typeId='%ld'\n",
		 childNode->format,(long)childNode->typeId);
      goto fail;
   }

   if(childNode->ndims==0) {
      if(writeScalarDataAttribute(loc_id, childNode->typeId, childName, childNode->data)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to write scalar data attribute '%s'\n",childNode->name);
	 goto fail;
      }
   } else {
      if(writeSimpleDataAttribute(loc_id,childNode->typeId, childName,
				  childNode->ndims, childNode->dims,
				  childNode->data)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to write simple data attribute '%s'\n",childNode->name);
	 goto fail;
      }
   }
   status = 1;

   childNode->mark = NMARK_ORIGINAL;
   
fail:
   if(loc_id>=0) {
      if(parentType==GROUP_ID) {
	 H5Gclose(loc_id);
      } else if(parentType==DATASET_ID) {
	 H5Dclose(loc_id);
      } else {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not determine type of loc_id, could not close\n");
      }
   }
   
   return status;
}

int doAppendHdf5Group(hid_t file_id, HL_Node* parentNode, char* parentName,
		      HL_Node* childNode, char* childName)
{
   hid_t loc_id = -1;
   hid_t new_id = -1;
   int status = 0;
   
   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 group since either parentName or childName is NULL\n");
      return 0;
   }
   if(strcmp(parentName,"")==0) {
      if((loc_id = H5Gopen(file_id,"/"))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not open root group when reading attr '%s'\n",
		    childName);
	 goto fail;
      }
   } else {
      if((loc_id = H5Gopen(file_id,parentName))<0) {
	 if(_debug_hdf)
	    fprintf(stderr, "Could not open group '%s' when creating new group.\n",
		    parentName);
	 goto fail;
      }
   }

   if((new_id = H5Gcreate(loc_id,childName,0))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create new group %s\n",childNode->name);
      goto fail;
   }
   status = 1;
   childNode->mark = NMARK_ORIGINAL;
fail:
   if(loc_id>=0) {
      H5Gclose(loc_id);
   }
   if(new_id>=0) {
      H5Gclose(new_id);
   }
   return status;
}

int doAppendHdf5Dataset(hid_t file_id, HL_Node* parentNode, char* parentName,
			HL_Node* childNode, char* childName, int doCompress)
{
   hid_t loc_id=-1;
   hid_t new_id=-1;
   int status = 0;
   
   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 dataset since either parentName or childName is NULL\n");
      return 0;
   }
   if(strcmp(parentName,"")==0) {
      if((loc_id = H5Gopen(file_id,"/"))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Could not open root group when creating new dataset '%s'\n",
		    childName);
	 goto fail;
      }
   } else {
      if((loc_id = H5Gopen(file_id,parentName))<0) {
	 if(_debug_hdf)
	    fprintf(stderr, "Could not open group '%s' when creating new dataset.\n",
		    parentName);
	 goto fail;
      }
   }

   if(strcmp(childNode->format,DATAFORMAT_UNDEFINED)!=0 && childNode->typeId<0) {
      childNode->typeId = translateCharToDatatype(childNode->format);
   }
   if(childNode->typeId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't recognize datatype '%s' or typeId='%ld'\n",
		 childNode->format,(long)childNode->typeId);
      goto fail;
   }
   new_id = createSimpleDataset(loc_id, childNode->typeId,
				childName, childNode->ndims,
				childNode->dims,
				childNode->data, doCompress);
   if(new_id<0) {	
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create dataset %s\n",childNode->name);
      goto fail;
   }

   status = 1;
   childNode->mark = NMARK_ORIGINAL;
fail:
   if(loc_id>=0) {
      H5Gclose(loc_id);
   }
   if(new_id>=0) {
      H5Dclose(new_id);
   }
   return status;
}


int updateNodeList(HL_NodeList* nodelist, int doCompress)
{
   int i;
   HL_Node* parentNode=NULL;
   char parentName[256];
   char childName[256];
   hid_t file_id=-1;
   hid_t gid=-1;
   hid_t strhid=-1;

   if((file_id=openHlHdfFile(nodelist->filename,"rw"))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to open file %s\n",nodelist->filename);
      return 0;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      /*printf("Checking node '%s'\n",nodelist->nodes[i]->name);*/
      if(nodelist->nodes[i]->mark == NMARK_CREATED) {
	 /*printf("%s has been marked as created\n",nodelist->nodes[i]->name);*/
	 if(!extractParentChildName(nodelist->nodes[i],parentName,childName)) {
	    if(_debug_hdf)
	       fprintf(stderr,"Failed to extract parent, child name\n");
	    goto fail;
	 }
	 /*printf("Trying to locate parent\n");*/
	 if((strcmp(parentName,"")!=0) && !(parentNode = getNode(nodelist,parentName))) {
	    if(_debug_hdf)
	       fprintf(stderr,"Failed to locate parent node '%s'\n",parentName);
	    goto fail;
	 }

	 /*printf("Checking node type\n");*/
	 switch(nodelist->nodes[i]->type) {
	 case ATTRIBUTE_ID: {
	    /*printf("Appending attribute node\n");*/
	    if(!doAppendHdf5Attribute(file_id,parentNode,parentName,nodelist->nodes[i],childName)) {
	       goto fail;
	    }
	    break;
	 }
	 case GROUP_ID: {
	    if(!doAppendHdf5Group(file_id,parentNode, parentName, nodelist->nodes[i], childName)) { 
	       goto fail;
	    }	
	    break;
	 }
	 case DATASET_ID: {
	    if(!doAppendHdf5Dataset(file_id,parentNode, parentName, nodelist->nodes[i],childName,doCompress)) {
	       goto fail;
	    }
	    break;
	 }
	 case TYPE_ID: {
	    if(!doCommitHdf5Datatype(file_id,parentNode, parentName, nodelist->nodes[i],childName))
	       goto fail;
	    break;
	 }
	 default: {
	    fprintf(stderr,"Unrecognized type\n");
	    break;
	 }
	 }
      }
   }
   
   if(gid>=0)
      H5Gclose(gid);
   if(file_id>=0)
      H5Fclose(file_id);
   if(strhid>=0)
      H5Tclose(strhid);
   return 1;   
fail:
   if(gid>=0)
      H5Gclose(gid);
   if(file_id>=0)
      H5Fclose(file_id);
   if(strhid>=0)
      H5Tclose(strhid);
   return 0;	 
}
