/* --------------------------------------------------------------------------
$Id: write_hlhdf.c,v 1.1.1.1 2001/12/14 12:02:38 ahenja Exp $
Program:        

Description:    Common functions for writing an HDF5 file.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: write_hlhdf.c,v $
Revision 1.1.1.1  2001/12/14 12:02:38  ahenja
First CVS version of hlhdf

Revision 1.6  2001/12/07 13:20:32  ahenja
Updated with possibilities to read and write references.

Revision 1.5  2001/11/09 08:30:56  ahenja
Updated with possibility to write references, unfortunately i havent got the time to implement read support.

Revision 1.4  2001/11/08 12:27:20  adybbroe
Updated.

Revision 1.3  2001/04/26 14:16:19  adybbroe
Solved for version 1.4 incompatibility.

Revision 1.2  2000/03/22 08:41:33  ahenja
Updated with spell checking and header information.

-----------------------------------------------------------------------------*/
#include "hlhdf.h"

/************************************************
 * commitType
 ***********************************************/
herr_t commitType(hid_t loc_id, const char* name, hid_t type_id)
{
   return H5Tcommit(loc_id,name,type_id);
}

/************************************************
 * createStringType
 ***********************************************/
hid_t createStringType(size_t length)
{
   hid_t type;
   type=H5Tcopy(H5T_C_S1);
   H5Tset_size(type,length);
   return type;
}
/************************************************
 * setTypeSize
 ***********************************************/
herr_t setTypeSize(hid_t type_id,size_t theSize)
{
   return H5Tset_size(type_id,theSize);
}

/************************************************
 * closeType
 ***********************************************/
herr_t closeType(hid_t type_id)
{
   return H5Tclose(type_id);
}

/************************************************
 * writeScalarDataAttribute
 ***********************************************/
herr_t writeScalarDataAttribute(hid_t loc_id, hid_t type_id, const char* name, void* buf)
{
   hid_t aid=-1;
   hid_t attr_id=-1;
   herr_t status = 0;
   if((aid = H5Screate(H5S_SCALAR))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create scalar data space\n");
      status = -1;
      goto done;
   }
   if((attr_id=H5Acreate(loc_id,name,type_id,aid,H5P_DEFAULT))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create scalar attribute\n");
      status = -1;
      goto done;
   }

   if(H5Awrite(attr_id,type_id,buf)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to write scalar data to file\n");
      status = -1;
   }

 done:
   if(aid!=-1)
      H5Sclose(aid);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   return status;
}

/************************************************
 * writeScalarDataAttribute_fmt
 ***********************************************/
herr_t writeScalarDataAttribute_fmt(hid_t loc_id, const char* fmt, const char* name, void* buf)
{
   hid_t typeId=-1;
   herr_t status=0;
   if((typeId=translateCharToDatatype(fmt))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to translate from format to HDF type\n");
      return -1;
   }
   status = writeScalarDataAttribute(loc_id,typeId,name,buf);
   H5Tclose(typeId);
   return status;
}

/************************************************
 * writeSimpleDataAttribute
 ***********************************************/
herr_t writeSimpleDataAttribute(hid_t loc_id, hid_t type_id, const char* name, int ndims, hsize_t* dims, void* buf)
{
   hid_t attr_id=-1;
   hid_t dataspace=-1;
   herr_t status = 0;
   if((dataspace = H5Screate_simple(ndims,dims,NULL))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create simple dataspace for attribute\n");
      status = -1;
      goto done;
   }

   if((attr_id=H5Acreate(loc_id,name,type_id,dataspace,H5P_DEFAULT))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create simple attribute\n");
      status = -1;
      goto done;
   }

   if(H5Awrite(attr_id,type_id,buf)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to write simple data attribute to file\n");
      status = -1;
   }

 done:
   if(dataspace!=-1)
      H5Sclose(dataspace);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   return status;
}

/************************************************
 * writeSimpleDataAttribute_fmt
 ***********************************************/
herr_t writeSimpleDataAttribute_fmt(hid_t loc_id, const char* fmt, const char* name,
				    int ndims, hsize_t* dims, void* buf)
{
   hid_t typeId=-1;
   herr_t status=0;
   if((typeId=translateCharToDatatype(fmt))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to translate from format to HDF type\n");
      return -1;
   }
   status = writeSimpleDataAttribute(loc_id,typeId,name,ndims,dims,buf);
   H5Tclose(typeId);
   return status;
}

/************************************************
 * createSimpleDataset
 ***********************************************/
hid_t createSimpleDataset(hid_t loc_id, hid_t type_id, const char* name, int ndims, hsize_t* dims, void* buf,int compress)
{
   hid_t dataset=-1;
   hid_t dataspace=-1;
   hid_t props=-1;

   if((dataspace=H5Screate_simple(ndims,dims,NULL))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create simple dataspace for dataset\n");
      goto done;
   }

   if(compress>0 && compress<=9) {
      if((props = H5Pcreate(H5P_DATASET_CREATE))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to create the compression property\n");
	 goto done;
      }
      if(H5Pset_chunk(props,ndims,dims)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to set chunk size\n");
	 goto done;
      }
      if(H5Pset_deflate(props,compress)) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to set z compression to level 6\n");
	 goto done;
      }
      if((dataset=H5Dcreate(loc_id,name,type_id,dataspace,props))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to create the dataset\n");
	 goto done;
      }
   } else {
      if((dataset=H5Dcreate(loc_id,name,type_id,dataspace,H5P_DEFAULT))<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to create the dataset\n");
	 goto done;
      }
   }

   if(buf!=NULL) {
      if(H5Dwrite(dataset,type_id,H5S_ALL,H5S_ALL,H5P_DEFAULT,buf)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to write dataset\n");
	 goto done;
      }
   }

 done:
   if(dataspace!=-1)
      H5Sclose(dataspace);
   if(props!=-1)
      H5Pclose(props);

   return dataset;
}

/************************************************
 * createSimpleDataset_fmt
 ***********************************************/
hid_t createSimpleDataset_fmt(hid_t loc_id, const char* fmt, const char* name, int ndims, hsize_t* dims, void* buf, int compress)
{
   hid_t typeId=-1;
   hid_t dset=0;
   if((typeId=translateCharToDatatype(fmt))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to translate from format to HDF type\n");
      return -1;
   }
   dset = createSimpleDataset(loc_id,typeId,name,ndims,dims,buf,compress);
   H5Tclose(typeId);
   return dset;
}

/************************************************
 * closeDataset
 ***********************************************/
herr_t closeDataset(hid_t loc_id)
{
   return H5Dclose(loc_id);
}


/************************************************
 * createCompoundType
 ***********************************************/
hid_t createCompoundType(size_t size)
{
   return H5Tcreate(H5T_COMPOUND,size);
}

/************************************************
 * addAttributeToCompoundType
 ***********************************************/
herr_t addAttributeToCompoundType(hid_t loc_id, const char* name, size_t offset,hid_t type_id)
{
   return H5Tinsert(loc_id,name,offset,type_id);
}

/************************************************
 * addAttributeToCompoundType_fmt
 ***********************************************/
herr_t addAttributeToCompoundType_fmt(hid_t loc_id, const char* name, size_t offset,const char* fmt)
{
   herr_t status;
   hid_t type_id = translateCharToDatatype(fmt);
   if(type_id<0)
      return -1;
   status = addAttributeToCompoundType(loc_id,name,offset,type_id);
   H5Tclose(type_id);
   return status;
}

/************************************************
 * addArrayToCompoundType
 ***********************************************/
herr_t addArrayToCompoundType(hid_t loc_id, const char* name, size_t offset, int ndims,size_t* dims,hid_t type_id)
{
   herr_t status;
   hid_t array_type;
   
#ifdef H5_VERS_MINOR
#if H5_VERS_MINOR>=4
   array_type=H5Tarray_create(type_id,ndims,(hsize_t *)dims,NULL);
   status=H5Tinsert(loc_id,name,offset,array_type);
   H5Tclose(array_type);
#else
   status=H5Tinsert_array(loc_id,name,offset,ndims,dims,NULL,type_id);
   /*
     Incompatible with version 1.4 and later. Adam Dybbroe, 2001-04-26
     The below should be in line with the latest version...
   */     
#endif
#else
   array_type=H5Tarray_create(type_id,ndims,(hsize_t *)dims,NULL);
   status=H5Tinsert(loc_id,name,offset,array_type);
   H5Tclose(array_type);
#endif
   
   return status;
}

/************************************************
 * addArrayToCompoundType_fmt
 ***********************************************/
herr_t addArrayToCompoundType_fmt(hid_t loc_id, const char* name, size_t offset, \
				  int ndims, size_t* dims, const char* fmt)
{
   hid_t type_id;
   herr_t status;
   type_id = translateCharToDatatype(fmt); 
   if(type_id<0)
      return -1;
   status = addArrayToCompoundType(loc_id,name,offset,ndims,dims,type_id);
   H5Tclose(type_id);

   return status;
}


/************************************************
 * createGroup
 ***********************************************/
hid_t createGroup(hid_t loc_id, const char* groupname,const char* comment)
{
   hid_t group_id;
   if((group_id=H5Gcreate(loc_id,groupname,0))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create group\n");
      return -1;
   }
   if(comment) {
      if(H5Gset_comment(loc_id,groupname,comment)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to set comment for group\n");
      }
   }
   return group_id;
}

/************************************************
 * closeGroup
 ***********************************************/
herr_t closeGroup(hid_t group_id)
{
   return H5Gclose(group_id);
}

/************************************************
 * createReference
 ***********************************************/
herr_t createReference(hid_t loc_id, hid_t file_id, const char* name,
		       const char* targetname)
{
   hid_t aid=-1;
   hid_t attr_id=-1;
   hid_t attr_type=-1;
   hobj_ref_t  ref;
   herr_t status = 0;
   
   if((aid = H5Screate(H5S_SCALAR))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create scalar data space\n");
      status = -1;
      goto done;
   }

   if((attr_type=H5Tcopy(H5T_STD_REF_OBJ))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to copy H5T reference type\n");
      status = -1;
      goto done;
   }
   
   if((attr_id=H5Acreate(loc_id,name,attr_type,aid,H5P_DEFAULT))<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create scalar attribute\n");
      status = -1;
      goto done;
   }

   /* Of course I assume that the dataset is accurately specified
    * by targetname, otherwise well, implement a check here, will
    * probably require that the in argument is the actual dataset id :-/
    */
   if(H5Rcreate(&ref, file_id, targetname, H5R_OBJECT, -1)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create reference object\n");
      status = -1;
      goto done;
   }
   if(H5Awrite(attr_id,attr_type,&ref)<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to write scalar data to file\n");
      status = -1;
   }

 done:
   if(aid!=-1)
      H5Sclose(aid);
   if(attr_id!=-1)
      H5Aclose(attr_id);
   if(attr_type!=-1)
      H5Tclose(attr_type);
   return status;
}
