/* --------------------------------------------------------------------------
$Id: pyhl.c,v 1.1.1.1 2001/12/14 12:02:42 ahenja Exp $
Program:        

Description:    The pyhl module source code, provides a python interface
		to the HighLevel protocol for writing and reading
		HDF5 files.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: pyhl.c,v $
Revision 1.1.1.1  2001/12/14 12:02:42  ahenja
First CVS version of hlhdf

Revision 1.7  2001/11/09 11:53:16  adybbroe
removed printf statement.

Revision 1.6  2001/11/08 12:07:14  ahenja
Basically done nothing, just some cleanup.

 * Revision 1.5  2001/01/18  12:47:49  ahenja
 * Added lots of changes.
 *
Revision 1.4  2000/08/21 08:44:30  dmichels
Replaced PyArray_FromDimsAndData and
removed trailing \OOO character when reading strings.

Revision 1.3  2000/04/11 10:08:41  ahenja
Added functionality for handle lists of ints, floats, doubles, longs and strings.

Revision 1.2  2000/03/22 08:59:06  ahenja
Updated with spell checking and header information.
	
-----------------------------------------------------------------------------*/
#include <Python.h>
#include <arrayobject.h>

#include "pyhlhdf_common.h"
#include "read_vhlhdf.h"
#include "write_vhlhdf.h"
#include "update_vhlhdf.h"

static PyObject *ErrorObject;

/* Some defines for error handling */
#define raiseException(type,msg)\
{PyErr_SetString(type,msg);return NULL;}

#define setException(type,msg)\
{PyErr_SetString(type,msg);}

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_NodeList* nodelist;
} PyhlObject;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_Node* node;
} PyhlNode;

/*
 *Define that the PyhlObject is of the type Pyhl_Type
 */
staticforward PyTypeObject Pyhl_Type;
staticforward PyTypeObject PyhlNode_Type;

#define Pyhl_Check(op) ((op)->ob_type == &Pyhl_Type)
#define PyhlNode_Check(op) ((op)->ob_type == &PyhlNode_Type)

/* Pyhl module functions */
static void _dealloc(PyhlObject* pyhl_object)
{
   /*Nothing yet*/
   if(!pyhl_object)
      return;

   freeHL_NodeList(pyhl_object->nodelist);

   PyMem_DEL(pyhl_object);
}

static void _dealloc_pyhlnode(PyhlNode* pyhl_node)
{
   /*Nothing yet*/
   if(!pyhl_node)
      return;

   freeHL_Node(pyhl_node->node);

   PyMem_DEL(pyhl_node);
}

static PyObject* _pyhl_new_nodelist(PyObject* self, PyObject* args)
{
   PyhlObject* retv=NULL;
   retv=PyObject_NEW(PyhlObject,&Pyhl_Type);
   if(!retv)
      return NULL;

   if(!(retv->nodelist=newHL_NodeList())) {
      setException(PyExc_MemoryError,"Failed to create HL NodeList\n");
      _dealloc(retv);
      retv=NULL;
   }

   return (PyObject*)retv;
}

static PyObject* _pyhl_new_node(PyObject* self, PyObject* args)
{
   PyhlNode* retv=NULL;
   HL_Type type;
   char* nodename;

   if(!PyArg_ParseTuple(args,"is",&type,&nodename))
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
   case GROUP_ID:
   case DATASET_ID:
   case TYPE_ID:
   case REFERENCE_ID:
      break;
   default:
      printf("Unsupported node type %d\n",type);
      return NULL;
   }

   retv=PyObject_NEW(PyhlNode,&PyhlNode_Type);
   if(!retv)
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
      retv->node = newHL_Attribute(nodename);
      break;
   case GROUP_ID:
      retv->node = newHL_Group(nodename);
      break;
   case DATASET_ID:
      retv->node = newHL_Dataset(nodename);
      break;
   case TYPE_ID:
      retv->node = newHL_Datatype(nodename);
      break;
   case REFERENCE_ID:
      retv->node = newHL_Reference(nodename);
      break;
   default:
      retv->node = NULL;
      break;
   }

   if(!(retv->node)) {
      setException(PyExc_MemoryError,"Failed to create Node\n");
      _dealloc_pyhlnode(retv);
      retv=NULL;
   }

   return (PyObject*)retv;
}

static PyObject* _pyhl_read_nodelist(PyObject* self, PyObject* args)
{
   HL_NodeList* nodelist=NULL;
   PyhlObject* retv=NULL;
   char* filename=NULL;
   char* frompath=NULL;
   
   if(!PyArg_ParseTuple(args,"s|s",&filename,&frompath))
      return NULL;

   if(!frompath) {
      nodelist = readHL_NodeList(filename);
   } else {
      nodelist = readHL_NodeListFrom(filename,frompath);
   }
   if(!nodelist) {
      char errmsg[256];
      if(!frompath)
	 sprintf("Could not read file '%s'",filename);
      else
	 sprintf("Could not read file '%s' from '%s'",filename,frompath);
      setException(PyExc_IOError,errmsg);
      goto fail;
   }

   if(!(retv=(PyhlObject*)_pyhl_new_nodelist(NULL,NULL))) {
      setException(PyExc_MemoryError,"Could not allocate nodelist instance");
      goto fail;
   }

   /*Change the nodelist*/
   freeHL_NodeList(retv->nodelist);
   retv->nodelist=nodelist;
   return (PyObject*)retv;
 fail:
   freeHL_NodeList(nodelist);
   if(retv)
      _dealloc(retv);
   return NULL;
}

static PyObject* _pyhl_is_file_hdf5(PyObject* self, PyObject* args)
{
   char* filename;
   int val;

   if(!(PyArg_ParseTuple(args,"s",&filename)))
      return NULL;

   val = isHdf5File(filename);

   return PyInt_FromLong((int)val);
}

/* PyhlObject member methods */
static PyObject* _pyhl_add_node(PyhlObject* self, PyObject* args)
{
   PyObject* inp;
   PyhlNode* pyhlNode;
   HL_Node* aNode;

   if(!PyArg_ParseTuple(args,"O",&inp))
      return NULL;

   if(!PyhlNode_Check(inp)) {
      setException(PyExc_TypeError,"Trying to add node, which not is of PyhlNodeCore type");
      return NULL;
   }

   pyhlNode=(PyhlNode*)inp;
   aNode=pyhlNode->node;

   if(!addNode(self->nodelist,aNode)) {
      setException(PyExc_MemoryError,"Could not add node to nodelist");
      freeHL_Node(aNode);
      return NULL;
   }

   pyhlNode->node=NULL;

   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_write(PyhlObject* self, PyObject* args)
{
   char* filename;
   int doCompress=6;
   if(!PyArg_ParseTuple(args,"s|i",&filename,&doCompress))
      return NULL;
   strcpy(self->nodelist->filename,filename);
   if(!writeNodeList(self->nodelist,doCompress)) {
      setException(PyExc_IOError,"Could not write file");
      return NULL;
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_update(PyhlObject* self, PyObject* args)
{
   int doCompress=6;
   if(!PyArg_ParseTuple(args,"|i",&doCompress))
      return NULL;   
   if(!updateNodeList(self->nodelist,doCompress)) {
      setException(PyExc_IOError,"Could not update file");
      return NULL;
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_get_node_names(PyhlObject* self, PyObject* args)
{
   PyObject* retv=NULL;
   PyObject* pyo=NULL;
   int i;
   char errbuf[256];
   if(!(retv=PyDict_New())) {
      setException(PyExc_MemoryError,"Could not allocate dictionary");
      return NULL;
   }
   for(i=0;i<self->nodelist->nNodes;i++) {
      switch(self->nodelist->nodes[i]->type) {
      case ATTRIBUTE_ID:
	 pyo=PyInt_FromLong(ATTRIBUTE_ID);
	 break;
      case DATASET_ID:
	 pyo=PyInt_FromLong(DATASET_ID);
	 break;
      case TYPE_ID:
	 pyo=PyInt_FromLong(TYPE_ID);
	 break;
      case GROUP_ID:
	 pyo=PyInt_FromLong(GROUP_ID);
	 break;
      case REFERENCE_ID:
	 pyo=PyInt_FromLong(REFERENCE_ID);
	 break;
      default:
	 sprintf(errbuf,"Unknown type for node '%s': '%d'",self->nodelist->nodes[i]->name,
		 self->nodelist->nodes[i]->type);
	 setException(PyExc_TypeError,errbuf);
	 goto fail;
      }
      if(PyDict_SetItemString(retv,self->nodelist->nodes[i]->name,pyo)==-1) {
	 setException(PyExc_AttributeError,"Failed to set dictionary item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }

   return retv;
 fail:
   if(pyo)
      Py_XDECREF(pyo);
   if(retv)
      Py_XDECREF(retv);
   return NULL;
}

static PyObject* _pyhl_select_all(PyhlObject* self, PyObject* args)
{
   selectAllNodes(self->nodelist);
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_select_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;
   if(!selectNode(self->nodelist,nodename)) {
      sprintf(errbuf,"Could not select node '%s'",nodename);
      setException(PyExc_AttributeError,errbuf);
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_fetch(PyhlObject* self, PyObject* args)
{
   if(!fetchMarkedNodes(self->nodelist)) {
      setException(PyExc_IOError,"Could not fetch selected nodes");
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_get_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];
   HL_Node* node=NULL;
   PyhlNode* retv=NULL;
   PyObject* myArgs=NULL;
   int i;
   size_t npts;

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;

   if(!(node=getNode(self->nodelist,nodename))) {
      sprintf(errbuf,"Could not get node '%s'",nodename);
      setException(PyExc_IOError,errbuf);
      goto fail;
   }

   if(!(myArgs=Py_BuildValue("(is)",node->type,node->name))) {
      setException(PyExc_AttributeError,"Could not create argument tuple for allocating node");
      goto fail;
   }

   if(!(retv=(PyhlNode*)_pyhl_new_node((PyObject*)self,myArgs))) {
      setException(PyExc_AttributeError,"Could not create node to return");
      goto fail;
   }

   freeHL_Node(retv->node);

   retv->node=copyHL_Node(node);
   
   Py_XDECREF(myArgs);
   return (PyObject*)retv;
 fail:
   Py_XDECREF(myArgs);
   if(retv)
      _dealloc_pyhlnode(retv);
   return NULL;
}

/* PyhlNode member methods */
static PyObject* _pyhl_node_set_scalar_value(PyhlNode* self, PyObject* args)
{
   int itemSize;
   char* typename;
   PyObject* data;
   hid_t lhid=-1;
   hid_t tmpHid=-1;
   int chkVal=0;

   if(!PyArg_ParseTuple(args,"iOsl",&itemSize,&data,&typename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(strcmp(typename,"string")!=0 &&
      strcmp(typename,"compound")!=0 &&
      (tmpHid=translateCharToDatatype(typename))==-1) {
      char errmsg[256];
      sprintf(errmsg,"No datatype called '%s'.",typename);
      setException(PyExc_AttributeError,errmsg);
      return NULL;
   }

   if(tmpHid!=-1) {
      /* Well, ordinary attribute, just make sure the space is enough */
      size_t typeSize=H5Tget_size(tmpHid);
      chkVal=0;
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(short)) {
	    short mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(int)) {
	    int mData = (int)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(long)) {
	    long mData = (long)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else {
	    long long mData = (long long)PyLong_AsLongLong(data);
	    if(PyErr_Occurred()) {
	       setException(PyExc_AttributeError,"Could not translate to long long, probably not supported");
	       goto fail;
	    }
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float mData = (float)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if (typeSize<=sizeof(double)) {
	    double mData = (double)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      }
      H5Tclose(tmpHid);
      tmpHid=-1;
      if(!chkVal) {
	 setException(PyExc_TypeError,"Could not set scalar value");
	 goto fail;
      }
   } else if(strcmp(typename,"string")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,strlen(mData)+1,(unsigned char*)mData,typename,-1)) {
	 setException(PyExc_ValueError,"Could not set scalar string");
	 goto fail;
      }
   } else if(strcmp(typename,"compound")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,itemSize,(unsigned char*)mData,typename,lhid)) {
	 setException(PyExc_ValueError,"Could not set scalar compound value");
	 goto fail;
      }
   }

   if(tmpHid>=0)
      H5Tclose(tmpHid);
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(tmpHid>=0)
      H5Tclose(tmpHid);

   return NULL;
}

static PyObject* _pyhl_node_set_array_value(PyhlNode* self, PyObject* args)
{
   PyObject* pydims;
   PyObject* data;
   hid_t lhid=-1;
   char* typename;
   int itemSize;
   int n;
   hsize_t dims[4];
   int i;
   int ndim;
   PyObject* pyo;
   size_t tmpSize;

   if(!PyArg_ParseTuple(args,"iOOsl",&itemSize,&pydims,&data,&typename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(PyObject_Length(pydims)>4) {
      setException(PyExc_ValueError,"Not more than 4 dimensions are allowed");
      goto fail;
   }

   if(PyArray_Check(data) && isFormatSupported(typename)) {
      /*Ahh plain array*/
      if(PyObject_Length(pydims)!=((PyArrayObject*)data)->nd) {
	 setException(PyExc_ValueError,"Rank of Array != no of dims");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) { /*Extract dimensions*/
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
	 if(dims[i]!=((PyArrayObject*)data)->dimensions[i]) {
	    setException(PyExc_AttributeError,"Array dimensions != list dimensions");
	    goto fail;
	 }
      }
      if((tmpSize = whatSizeIsHdfFormat(typename))<0) {
	 setException(PyExc_ValueError,"Could not determine size");
	 goto fail;
      }
      if(tmpSize!=((PyArrayObject*)data)->descr->elsize) {
	 setException(PyExc_ValueError,"Type sizes are different between format and array");
	 goto fail;
      }
      if(!setArrayValue(self->node,tmpSize,ndim,dims,(unsigned char*)((PyArrayObject*)data)->data,
			typename,-1)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else if(PyList_Check(data) && (strcmp(typename,"string")==0 || isFormatSupported(typename))) {
      /* Okie, list of something */
      int maxstrlen=0;
      char* tmpData=NULL;
      hid_t strtype=-1;

      if(PyObject_Length(pydims)!=1) {
	 setException(PyExc_ValueError,"When using lists, the only allowed rank is 1.");
	 goto fail;
      }
      n=PyObject_Length(data);
      if(strcmp(typename,"string")==0) {
	 /*Calculate the longest string*/
	 for(i=0;i<n;i++) {
	    char* tmpstr;
	    if(!(pyo=PySequence_GetItem(data,i))) {
	       setException(PyExc_AttributeError,"Could not get list item");
	       goto fail;
	    }
	    if(!(tmpstr = PyString_AsString(pyo))) {
	       setException(PyExc_AttributeError,"Could not translate data to a string");
	       goto fail;
	    }
	    if(strlen(tmpstr)>maxstrlen)
	       maxstrlen=strlen(tmpstr);
	    Py_XDECREF(pyo);
	    pyo=NULL;
	 }
	 itemSize = maxstrlen + 1;
	 if(!(tmpData=malloc(itemSize * n))) {
	    setException(PyExc_MemoryError,"Could not allocate memory for strings");
	    goto fail;
	 }
	 memset(tmpData,0,itemSize*n);
	 for(i=0;i<n;i++) {
	    char* tmpstr;
	    pyo = PySequence_GetItem(data,i);
	    tmpstr = PyString_AsString(pyo);
	    strncpy(&tmpData[i*itemSize],tmpstr,strlen(tmpstr));
	 }
	 if((strtype = createStringType(itemSize))<0) {
	    setException(PyExc_MemoryError,"Could not create a string type");
	    free(tmpData);
	    goto fail;
	 }
	 dims[0]=n;
	 if(!(setArrayValue(self->node,itemSize,1,dims,(unsigned char*)tmpData,typename,strtype))) {
	    setException(PyExc_AttributeError,"Could not set array data");
	    H5Tclose(strtype);
	    free(tmpData);
	    goto fail;
	 }
      } else {
	 double* doubleptr=NULL;
	 float* floatptr=NULL;
	 int* intptr=NULL;
	 long* longptr=NULL;
	 if((tmpSize = whatSizeIsHdfFormat(typename))<0) {
	    setException(PyExc_ValueError,"Could not determine size");
	    goto fail;
	 }
	 if(strcmp(typename,"double")==0) {
	    doubleptr = malloc(sizeof(double)*n);
	    tmpData = (char*)doubleptr;
	 } else if(strcmp(typename,"float")==0) {
	    floatptr = malloc(sizeof(float)*n);
	    tmpData = (char*)floatptr;
	 } else if(strcmp(typename,"int")==0) {
	    intptr = malloc(sizeof(int)*n);
	    tmpData = (char*)intptr;
	 } else if(strcmp(typename,"long")==0) {
	    longptr = malloc(sizeof(long)*n);
	    tmpData = (char*)longptr;
	 } else {
	    setException(PyExc_TypeError,"Only supported types on list are double,float,int and long");
	    goto fail;
	 }
	 if(!tmpData) {
	    setException(PyExc_MemoryError,"Could not allocate memory for datatype");
	    goto fail;
	 }

	 for(i=0;i<n;i++) {
	    if(!(pyo=PySequence_GetItem(data,i))) {
	       setException(PyExc_AttributeError,"Could not get list item");
	       free(tmpData);
	       goto fail;
	    }
	    if(doubleptr) {
	       doubleptr[i] = (double)PyFloat_AsDouble(pyo);
	    } else if(floatptr) {
	       floatptr[i] = (float)PyFloat_AsDouble(pyo);
	    } else if(intptr) {
	       intptr[i] = (int)PyInt_AsLong(pyo);
	    } else if(longptr) {
	       longptr[i] = (long)PyInt_AsLong(pyo);
	    } else {
	       setException(PyExc_AttributeError,"Whoa, something very strange happened");
	       free(tmpData);
	       goto fail;
	    }
	    Py_XDECREF(pyo);
	    pyo=NULL;
	 }
	 dims[0]=n;
	 if(!(setArrayValue(self->node,tmpSize,1,dims,(unsigned char*)tmpData,typename,-1))) {
	    setException(PyExc_AttributeError,"Could not set array data");
	    free(tmpData);
	    goto fail;
	 }
      }
      if(tmpData) free(tmpData);
   } else if(strcmp(typename,"compound")==0) {
      size_t tmpLen=0;
      char* tmpData=PyString_AsString(data);
      if(!tmpData) {
	 setException(PyExc_AttributeError,"Could not translate data to a string");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) {
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
      }
      tmpLen=1;
      for(i=0;i<ndim;i++) {
	 tmpLen*=dims[i];
      }
      tmpLen*=itemSize;
      if(PyObject_Length(data)!=tmpLen) {
	 setException(PyExc_AttributeError,"There is some inconsistency between data size and calculated datasize");
	 goto fail;
      }
      if(!setArrayValue(self->node,itemSize,ndim,dims,(unsigned char*)tmpData,typename,lhid)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else {
      setException(PyExc_AttributeError,"Only support arrays and compound types right now");
      goto fail;
   }

   if(pyo) Py_XDECREF(pyo);
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(pyo) Py_XDECREF(pyo);
   return NULL;
}

static PyObject* _pyhl_node_commit(PyhlNode* self, PyObject* args)
{
   hid_t dType;
   if(!PyArg_ParseTuple(args,"l",&dType))
      return NULL;
   if(!self->node->type==TYPE_ID) {
      setException(PyExc_AttributeError,"Trying to commit a node which not is a type node");
      return NULL;
   }
   commitDatatype(self->node,dType);
   Py_INCREF(Py_None);
   return Py_None;
}
   
static PyObject* _pyhl_node_name(PyhlNode* self, PyObject* args)
{
   return PyString_FromString(self->node->name);
}

static PyObject* _pyhl_node_type(PyhlNode* self, PyObject* args)
{
   return PyInt_FromLong(self->node->type);
}

static PyObject* _pyhl_node_dims(PyhlNode* self,PyObject* args)
{
   PyObject* retv=NULL;
   int i;
   PyObject* pyo=NULL;
   if(!(retv=PyList_New(0))) {
      return NULL;
   }
   for(i=0;i<self->node->ndims;i++) {
      if(!(pyo = PyInt_FromLong(self->node->dims[i]))) {
	 setException(PyExc_ValueError,"Could not create py integer");
	 goto fail;
      }
      if(PyList_Append(retv,pyo)==-1) {
	 setException(PyExc_ValueError,"Could not append list item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }
   Py_XDECREF(pyo);
   return retv;
 fail:
   Py_XDECREF(pyo);
   Py_XDECREF(retv);
   return NULL;
}

static PyObject* _pyhl_node_format(PyhlNode* self,PyObject* args)
{
   return PyString_FromString(self->node->format);
}

static PyObject* _pyhl_node_data(PyhlNode* self, PyObject* args)
{
   PyObject* retv=NULL;
   hid_t tmpHid=-1;
   size_t typeSize;
   char errbuf[256];
   int* dims=NULL;
   int i;
   size_t npts;

   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   typeSize = H5Tget_size(tmpHid);

   if(self->node->ndims==0) { /*Scalar*/
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(short)) {
	    short v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(int)) {
	    int v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(long)) {
	    long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else {
	    long long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyLong_FromLongLong(v);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble((double)v);
	 } else if(typeSize<=sizeof(double)) {
	    double v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble(v);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      case H5T_COMPOUND: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize);
	 break;
      }
      case H5T_STRING: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize-1);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   } else { /*Simple*/
      if(!(dims = malloc(sizeof(int)*self->node->ndims))) {
	 setException(PyExc_MemoryError,"Could not allocate dims");
	 goto fail;
      }
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER:
      case H5T_FLOAT: {
	 int iformat=pyarraytypeFromHdfType(self->node->format);
	 int nbytes;
	 if(iformat==-1) {
	    sprintf(errbuf,"Unrecognized datatype %s",self->node->format);
	    setException(PyExc_TypeError,errbuf);
	    goto fail;
	 }
	 for(i=0;i<self->node->ndims;i++)
	    dims[i]=(int)self->node->dims[i];

	 /* Replaced PyArray_FromDimsAndData with this */
	 retv = PyArray_FromDims(self->node->ndims,dims,iformat);
	 if(!retv) {
	    setException(PyExc_MemoryError,"Could not create array");
	    goto fail;
	 }
	 nbytes = 1;
	 for(i=0;i<self->node->ndims;i++)
	    nbytes *= dims[i];
	 nbytes *= ((PyArrayObject*)retv)->descr->elsize;

	 memcpy(((PyArrayObject*)retv)->data,(unsigned char*)self->node->data,nbytes);
	 break;
      }
      case H5T_COMPOUND: {
	 npts=1;
	 for(i=0;i<self->node->ndims;i++)
	    npts*=self->node->dims[i];
	 npts*=typeSize;
	 retv=PyString_FromStringAndSize((char*)self->node->data,npts);
	 break;
      }
      case H5T_STRING: {
	 if(self->node->ndims!=1) {
	    /* Dunno how to represent a multi-dim array of strings */
	    npts=1;
	    for(i=0;i<self->node->ndims;i++)
	       npts*=self->node->dims[i];
	    npts*=typeSize;
	    retv=PyString_FromStringAndSize((char*)self->node->data,npts-1);
	 } else {
	    retv = PyList_New(0);
	    for(i=0;retv && i<self->node->dims[0];i++) {
	       PyObject* pyo = PyString_FromStringAndSize((char*)(&self->node->data[i*typeSize]),typeSize-1);
	       if(!pyo) {
		  setException(PyExc_MemoryError,"Could not create string");
		  goto fail;
	       }
	       if(PyList_Append(retv,pyo)==-1) {
		  Py_XDECREF(pyo);
		  goto fail;
	       }
	       Py_XDECREF(pyo);
	    }
	 }
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   }
   if(dims)
      free(dims);
   return retv;
 fail:
   Py_XDECREF(retv);
   if(dims)
      free(dims);
   return NULL;
}

static PyObject* _pyhl_node_rawdata(PyhlNode* self, PyObject* args)
{
   PyObject* retv=NULL;
   hid_t tmpHid=-1;
   size_t typeSize;
   char errbuf[256];
   int* dims=NULL;
   int i;
   size_t npts;

   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   typeSize = self->node->dSize; /* Using raw type */

   if(self->node->ndims==0) { /*Scalar*/
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(short)) {
	    short v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(int)) {
	    int v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(long)) {
	    long v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else {
	    long long v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyLong_FromLongLong(v);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyFloat_FromDouble((double)v);
	 } else if(typeSize<=sizeof(double)) {
	    double v;
	    memcpy(&v,self->node->rawdata,typeSize);
	    retv=PyFloat_FromDouble(v);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      case H5T_COMPOUND: {
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,typeSize);
	 break;
      }
      case H5T_STRING: {
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,typeSize-1);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   } else { /*Simple*/
      if(!(dims = malloc(sizeof(int)*self->node->ndims))) {
	 setException(PyExc_MemoryError,"Could not allocate dims");
	 goto fail;
      }
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER:
      case H5T_FLOAT: {
	 int iformat=pyarraytypeFromHdfType(self->node->format);
	 int nbytes;
	 if(iformat==-1) {
	    sprintf(errbuf,"Unrecognized datatype %s",self->node->format);
	    setException(PyExc_TypeError,errbuf);
	    goto fail;
	 }
	 for(i=0;i<self->node->ndims;i++)
	    dims[i]=(int)self->node->dims[i];

	 /* Replaced PyArray_FromDimsAndData with this */
	 retv = PyArray_FromDims(self->node->ndims,dims,iformat);
	 if(!retv) {
	    setException(PyExc_MemoryError,"Could not create array");
	    goto fail;
	 }
	 nbytes = 1;
	 for(i=0;i<self->node->ndims;i++)
	    nbytes *= dims[i];
	 nbytes *= ((PyArrayObject*)retv)->descr->elsize;

	 memcpy(((PyArrayObject*)retv)->data,(unsigned char*)self->node->rawdata,nbytes);
	 break;
      }
      case H5T_COMPOUND: {
	 npts=1;
	 for(i=0;i<self->node->ndims;i++)
	    npts*=self->node->dims[i];
	 npts*=typeSize;
	 retv=PyString_FromStringAndSize((char*)self->node->rawdata,npts);
	 break;
      }
      case H5T_STRING: {
	 if(self->node->ndims!=1) {
	    /* Dunno how to represent a multi-dim array of strings */
	    npts=1;
	    for(i=0;i<self->node->ndims;i++)
	       npts*=self->node->dims[i];
	    npts*=typeSize;
	    retv=PyString_FromStringAndSize((char*)self->node->rawdata,npts-1);
	 } else {
	    retv = PyList_New(0);
	    for(i=0;retv && i<self->node->dims[0];i++) {
	       PyObject* pyo = PyString_FromStringAndSize((char*)(&self->node->rawdata[i*typeSize]),typeSize-1);
	       if(!pyo) {
		  setException(PyExc_MemoryError,"Could not create string");
		  goto fail;
	       }
	       if(PyList_Append(retv,pyo)==-1) {
		  Py_XDECREF(pyo);
		  goto fail;
	       }
	       Py_XDECREF(pyo);
	    }
	 }
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   }
   if(dims)
      free(dims);
   return retv;
 fail:
   Py_XDECREF(retv);
   if(dims)
      free(dims);
   return NULL;
}

static PyObject* getPythonObjectFromNode(HL_CompoundTypeAttribute* descr ,unsigned char* data, int idx)
{
   PyObject* pyo = NULL;

   int offset = descr->offset + idx * descr->size;
   int debug = 0;
   
   char tmpstr[300];
   memset(tmpstr,0,300);

   if(debug) {
      printf("name = %s, offset = %d, descr->offset = %d, idx = %d, descr->size=%d, ",descr->attrname, offset, descr->offset,
	     idx, descr->size);
   }

   if(strcmp(descr->format,"string")==0) {
      int len = strlen((char*)&data[offset]);
      char* dummystr = "";
      if( len > 0 && len <=descr->size) {
	 if(debug)
	    strncpy(&tmpstr[0],(char*)&data[offset],len);
	 pyo = PyString_FromStringAndSize((char*)&data[offset],len);
      } else if(len <= 0) {
	 if(debug)
	    strncpy(&tmpstr[0],dummystr,len);
	 pyo = PyString_FromString(dummystr);
      } else if(len >= descr->size) {
	 if(debug)
	    strncpy(&tmpstr[0],(char*)&data[offset],len);
	 pyo = PyString_FromStringAndSize((char*)&data[offset],descr->size);
      }
      if(debug)
	 printf("val = '%s'\n",tmpstr);
   } else if(strcmp(descr->format,"schar")==0 ||
	     strcmp(descr->format,"uchar")==0 ||
	     strcmp(descr->format,"char")==0) {
      char val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%c\n",val);
      pyo = PyString_FromStringAndSize(&val,1);
   } else if(strcmp(descr->format,"short")==0 ||
	     strcmp(descr->format,"ushort")==0) {
      short val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"int")==0 ||
	     strcmp(descr->format,"uint")==0) {
      int val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"long")==0 ||
	     strcmp(descr->format,"ulong")==0) {
      long val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"llong")==0) {
      long long val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"float")==0) {
      float val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%f\n",val); /*offset = descr->offset + idx * descr->size;*/
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"double")==0) {
      double val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("dval=%f\n",val);
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"ldouble")==0) {
      double val = 0.0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%f\n",val);
      pyo = PyFloat_FromDouble((double)val);
   } else if(strcmp(descr->format,"hsize")==0) {
      hsize_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"hssize")==0) {
      hssize_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"herr")==0) {
      herr_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else if(strcmp(descr->format,"hbool")==0) {
      hbool_t val = 0;
      memcpy((unsigned char*)&val,&data[offset],descr->size);
      if(debug)
	 printf("val=%d\n",val);
      pyo = PyInt_FromLong((long)val);
   } else {
      char errmsg[256];
      sprintf(errmsg,"Unsupported data type '%s'\n",descr->format);
      setException(PyExc_AttributeError,errmsg);
   }
   return pyo;
}

static PyObject* _pyhl_node_get_compound_data(PyhlNode* self, PyObject* args)
{
   hid_t tmpHid=-1;
   PyObject* retv=NULL;
   PyObject *pyo=NULL,*pyo2=NULL;
   char errbuff[256];
   int* dims=NULL;
   int i,j;
   size_t npts;
   unsigned char* data;
   HL_CompoundTypeDescription* descr;
   
   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   if(H5Tget_class(tmpHid)!=H5T_COMPOUND) {
      setException(PyExc_AttributeError,"Only supposed to be used with composite nodes");
      goto fail;
   }
   
   if(self->node->ndims==0) { /*Scalar*/
      if(!(retv=PyDict_New())) {
	 goto fail;
      }
      if(!self->node->compoundDescription) {
	 setException(PyExc_AttributeError,"Node does not have a compound description");
	 goto fail;
      }
      if(!(retv = PyDict_New())) {
	 setException(PyExc_MemoryError,"Could not allocate dicionary\n");
	 goto fail;
      }
      data = self->node->data;
      descr = self->node->compoundDescription;
      for(i=0;i<descr->nAttrs;i++) {
	 pyo = NULL;
	 if(descr->attrs[i]->ndims == 0) {
	    if(!(pyo = getPythonObjectFromNode(descr->attrs[i],data,0))) {
	       goto fail;
	    }
	 } else {
	    int ndim=1;
	    if(!(pyo = PyList_New(0))) {
	       goto fail;
	    }
	    for(j=0;j<descr->attrs[i]->ndims;j++) {
	       ndim *= descr->attrs[i]->dims[j];
	    }
	    /*
	     *if(strcmp(descr->attrs[i]->attrname,"area_extent")==0) {
	     *  int k = 0;
	     *  double tmpv = 0.0;
	     *  for(k=0;k<1342;k++) {
	     *	  tmpv = 0.0;
	     *	  memcpy((unsigned char*)&tmpv,&data[k],8);
	     *	  printf("tmpv(%d)=%f\n",k,tmpv);
	     *  }
	     *}
	     */
	    for(j=0;j<ndim;j++) {
	       if(!(pyo2 = getPythonObjectFromNode(descr->attrs[i],data,j))) {
		  goto fail;
	       }
	       PyList_Append(pyo,pyo2);
	       Py_XDECREF(pyo2);
	       pyo2 = NULL;
	    }
	 }

	 if(PyDict_SetItemString(retv,descr->attrs[i]->attrname,pyo)!=0) {
	    goto fail;
	 }
	 Py_XDECREF(pyo);
	 pyo = NULL;

	 /*
	 printf("attrname = %s\n",descr->attrs[i]->attrname);
	 printf("offset = %d\n",descr->attrs[i]->offset);
	 printf("ndims = %d\n",descr->attrs[i]->ndims);
	 printf("format = %s\n",descr->attrs[i]->format);
	 printf("size = %d\n",descr->attrs[i]->size);
	 for(j=0;j<descr->attrs[i]->ndims;j++)
	    printf("dim(%d)=%d\n",j,descr->attrs[i]->dims[j]);*/
      }
   } else {
      setException(PyExc_AttributeError,"Method does not support arrays of compound objects yet.");
      goto fail;
   }
   Py_XDECREF(pyo);
   Py_XDECREF(pyo2);
   
   return retv;
fail:
   Py_XDECREF(retv);
   Py_XDECREF(pyo);
   Py_XDECREF(pyo2);
   return NULL;
}

static struct PyMethodDef methods[] =
{
  {"addNode",(PyCFunction)_pyhl_add_node,1},
  {"write",(PyCFunction)_pyhl_write,1},
  {"update",(PyCFunction)_pyhl_update,1},
  {"getNodeNames",(PyCFunction)_pyhl_get_node_names,1},
  {"selectAll",(PyCFunction)_pyhl_select_all,1},
  {"selectNode",(PyCFunction)_pyhl_select_node,1},
  {"fetch",(PyCFunction)_pyhl_fetch,1},
  {"getNode",(PyCFunction)_pyhl_get_node,1},
  {NULL, NULL} /* sentinel */
};

static struct PyMethodDef node_methods[] =
{
   {"setScalarValue",(PyCFunction)_pyhl_node_set_scalar_value,1},
   {"setArrayValue",(PyCFunction)_pyhl_node_set_array_value,1},
   {"commit",(PyCFunction)_pyhl_node_commit,1},
   
   /* Inquiry options */
   {"name",(PyCFunction)_pyhl_node_name,1},
   {"type",(PyCFunction)_pyhl_node_type,1},
   {"dims",(PyCFunction)_pyhl_node_dims,1},
   {"format",(PyCFunction)_pyhl_node_format,1},
   {"data",(PyCFunction)_pyhl_node_data,1},
   {"rawdata",(PyCFunction)_pyhl_node_rawdata,1},
   {"compound_data",(PyCFunction)_pyhl_node_get_compound_data,1},
   {NULL, NULL} /* sentinel */
};

static PyObject* _getattr(PyhlObject* self, char* name)
{
    PyObject* res;

    if(strcmp(name,"ATTRIBUTE_ID")==0)
       return PyInt_FromLong(ATTRIBUTE_ID);
    if(strcmp(name,"GROUP_ID")==0)
       return PyInt_FromLong(GROUP_ID);
    if(strcmp(name,"DATASET_ID")==0)
       return PyInt_FromLong(DATASET_ID);
    if(strcmp(name,"TYPE_ID")==0)
       return PyInt_FromLong(TYPE_ID);
    if(strcmp(name,"REFERENCE_ID")==0)
       return PyInt_FromLong(REFERENCE_ID);
    
    res = Py_FindMethod(methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static PyObject* _getattr_node(PyhlNode* self, char* name)
{
    PyObject* res;

    res = Py_FindMethod(node_methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

statichere PyTypeObject Pyhl_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlCore",		/*tp_name*/
    sizeof(PyhlObject),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc,	/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject PyhlNode_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlNodeCore",		/*tp_name*/
    sizeof(PyhlNode),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_pyhlnode,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_node,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

static PyMethodDef functions[] = {
  {"nodelist", (PyCFunction)_pyhl_new_nodelist, 1},
  {"node", (PyCFunction)_pyhl_new_node, 1},
  {"read_nodelist",(PyCFunction)_pyhl_read_nodelist,1},
  {"is_file_hdf5",(PyCFunction)_pyhl_is_file_hdf5,1},
  {NULL,NULL} /*Sentinel*/
};

void init_pyhl()
{
   PyObject *m,*d;
   PyObject *tmp;
   Pyhl_Type.ob_type=&PyType_Type;
   PyhlNode_Type.ob_type=&PyType_Type;

   m=Py_InitModule("_pyhl",functions);
   d=PyModule_GetDict(m);

   tmp = PyInt_FromLong(ATTRIBUTE_ID);
   PyDict_SetItemString(d,"ATTRIBUTE_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(GROUP_ID);
   PyDict_SetItemString(d,"GROUP_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(DATASET_ID);
   PyDict_SetItemString(d,"DATASET_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(TYPE_ID);
   PyDict_SetItemString(d,"TYPE_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(REFERENCE_ID);
   PyDict_SetItemString(d,"REFERENCE_ID",tmp);
   Py_XDECREF(tmp);   
   
   import_array(); /*To make sure I get access to Numeric*/
   /*Always have to do this*/
   initHlHdf();
   /*And this I just do to be able to get debugging info from hdf*/
   debugHlHdf(2);
}
